#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <stdlib.h>
#include <inttypes.h>
#include "string.h"

#define BUFFER_LENGTH 16

/* UART Buffer Defines */
#define UART_RX_BUFFER_SIZE 16 /* 1,2,4,8,16,32,64,128 or 256 bytes */
#define UART_RX_BUFFER_MASK ( UART_RX_BUFFER_SIZE - 1 )
#if ( UART_RX_BUFFER_SIZE & UART_RX_BUFFER_MASK )
#error RX buffer size is not a power of 2
#endif
#define UART_TX_BUFFER_SIZE 16 /* 1,2,4,8,16,32,64,128 or 256 bytes */
#define UART_TX_BUFFER_MASK ( UART_TX_BUFFER_SIZE - 1 )

/* Static Variables */
static unsigned char UART_RxBuf[UART_RX_BUFFER_SIZE];
static volatile unsigned char UART_RxHead;
static volatile unsigned char UART_RxTail;
static unsigned char UART_TxBuf[UART_TX_BUFFER_SIZE];
static volatile unsigned char UART_TxHead;
static volatile unsigned char UART_TxTail;
static unsigned char ADCReady;

/* initialize UART */
void InitUART( unsigned char baudrate )
{
   unsigned char x;
   outp((unsigned char)(baudrate >> 8), UBRRH);
   outp((unsigned char)(baudrate), UBRRL);
   /* enable UART receiver and transmitter, and receive interrupt */
   outp((1<<RXCIE) | (1<<RXEN) | (1<<TXEN), UCSRB);
   outp((1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0), UCSRC);
   x = 0; /* flush receive buffer */
   UART_RxTail = x;
   UART_RxHead = x;
   UART_TxTail = x;
   UART_TxHead = x;
}

/* interrupt handlers */
//interrupt [UART_RX_vect] void UART_RX_interrupt( void )
SIGNAL(SIG_UART_RECV)
{
   unsigned char data;
   unsigned char tmphead;
   data = inp(UDR);
   /* calculate buffer index */
   tmphead = ( UART_RxHead + 1 ) & UART_RX_BUFFER_MASK;
   UART_RxHead = tmphead; /* store new index */
   if ( tmphead == UART_RxTail )
   {
      /* ERROR! Receive buffer overflow */
   }
   UART_RxBuf[tmphead] = data; /* store received data in buffer */
}

//interrupt [UART_UDRE_vect] void UART_TX_interrupt(void)
SIGNAL(SIG_UART_DATA)
{
   unsigned char tmptail;

   /* check if all data is transmitted */
   if ( UART_TxHead != UART_TxTail )
   {
      /* calculate buffer index */
      tmptail = ( UART_TxTail + 1 ) & UART_TX_BUFFER_MASK;
      UART_TxTail = tmptail; /* store new index */
      outp(UART_TxBuf[tmptail], UDR);
//      outp(0xAA, UDR);
   }
   else
   {
//      outp(inp(UCR) & ~(1 << UDRIE), UCR);
      cbi(UCSRB, UDRIE);
   }
}

SIGNAL(SIG_ADC)
{
   ADCReady = 1;
}

/* Read and write functions */
void TransmitByte( unsigned char data )
{
   unsigned char tmphead;
   /* calculate buffer index */
   tmphead = ( UART_TxHead + 1 ) & UART_TX_BUFFER_MASK; /* wait for free space in buffer */
   while ( tmphead == UART_TxTail );

   UART_TxBuf[tmphead] = data; /* store data in buffer */
   UART_TxHead = tmphead; /* store new index */
//   outp(inp(UCR) | (1 << UDRIE), UCR);
   sbi(UCSRB, UDRIE);
}

void TransmitLine(unsigned char *TransmitBuffer)
{
   unsigned char BufferCounter = 0;
   while (TransmitBuffer[BufferCounter] != '\0')
   {
      TransmitByte(TransmitBuffer[BufferCounter++]);
   }
}

unsigned char ReceiveByte( void )
{
   unsigned char tmptail;
   while ( UART_RxHead == UART_RxTail ) /* wait for incomming data */
   ;
   tmptail = ( UART_RxTail + 1 ) & UART_RX_BUFFER_MASK;/* calculate buffer index */
   UART_RxTail = tmptail; /* store new index */
   return UART_RxBuf[tmptail]; /* return data */
}

void ReceiveLine(unsigned char *ReturnBuffer) // ReturnBuffer must be at least BUFFER_LENGTH long
{
   unsigned char BufferCounter = 0;
//   unsigned char OutputLine[20];
//   unsigned char TempLine[20];
   while (BufferCounter < BUFFER_LENGTH - 2)
   {
      ReturnBuffer[BufferCounter] = ReceiveByte();
      if ((ReturnBuffer[BufferCounter] == '\n') || (ReturnBuffer[BufferCounter] == '\r'))
      {
         ReturnBuffer[BufferCounter] = '\0';
         break;
      }
      if (((ReturnBuffer[BufferCounter] == '\b') || (ReturnBuffer[BufferCounter] == 127)) && (BufferCounter > 0))
      {
         BufferCounter --;
         TransmitByte('\b');
         continue;
      }
      TransmitByte(ReturnBuffer[BufferCounter]);
      BufferCounter ++;
   }
   ReturnBuffer[BufferCounter++] = '\0';
}

unsigned char DataInReceiveBuffer(void)
{
   return(UART_RxHead != UART_RxTail); /* return 0 (FALSE) if the receive buffer is empty */
}

void clr(void)
{ // Send a 'clear screen' to a VT100 terminal
   TransmitByte(0x1B);
   TransmitByte('[');
   TransmitByte('H');
   TransmitByte(0x1B);
   TransmitByte('[');
   TransmitByte('2');
   TransmitByte('J');
   TransmitByte(0x00);
}

void SleepMode(char Idle)
{
   /* idle=0 - power down mode
    * idle=1 - idle mode
    * idle=2 - power save mode
   */
   if (Idle == 0) /* power down mode */
   {
      sbi(MCUCR, SM1);
      cbi(MCUCR, SM0);
   }
   else if (Idle == 2)
   {
      sbi(MCUCR, SM1);
      sbi(MCUCR, SM0);
   }
   else
   {
      cbi(MCUCR, SM1);
      cbi(MCUCR, SM0);
   }
   
   sbi(MCUCR, SE); /* enable sleep */
   asm volatile ("sleep"); /* idle mode */
}

int SetupADC(int Channel)
{
   outp(Channel, ADMUX); /* enable analog channel 0 */
   outp(0x88 | 0x07, ADCSR); /* ADC enable and set prescaler */
   sbi(ADCSR, 6); /* trigger dummy read */
   sei();
   return(0);
}

int StartADC(int Channel)
{
   int Result;

   ADCReady = 0;
   sbi(ADCSR, 6);

   SleepMode(1);
   Result = inp(ADCL);
   Result += inp(ADCH) << 8;
   return(Result);
}

/* main - a simple test program*/
int main( void )
{
   unsigned char InputChar;
   unsigned char TempBuffer[BUFFER_LENGTH];
   unsigned int ADCValue;
   unsigned char UartCounter;
//   float Temperature;
   unsigned long TemperatureInt;
   char Counter;
   
   InitUART(23);

   outp(0x80, DDRD); /* enable D7 as output */
   outp(inp(PORTD) | 0x80, PORTD);

   SetupADC(0);

   ADCReady = 0;
   ADCValue = StartADC(0);

   clr();
   TransmitLine("Starting up\n\r");
//   TransmitByte(sizeof(TemperatureInt) + '0');
   while ( 1 )
   {

      InputChar = ReceiveByte();

      outp(inp(PORTD) | 0x80, PORTD);
      TemperatureInt = 0;
      for (Counter = 0 ; Counter < 32 ; Counter ++)
      {
         ADCValue = StartADC(0);
	 TemperatureInt += ADCValue;
      }
      outp(inp(PORTD) & 0x3F, PORTD);
      TemperatureInt = TemperatureInt / 32;
      TemperatureInt = (TemperatureInt * 2500) / 1024;

      our_itoa(TemperatureInt / 10, TempBuffer);
      TransmitLine(TempBuffer);
      TransmitLine(".");
      our_itoa(TemperatureInt % 10, TempBuffer);
      TransmitLine(TempBuffer);
      TransmitLine(" degrees\n\r");
   }
}
