#ifndef __SERIAL_H__
#define __SERIAL_H__

#define UART_BUFFER_LENGTH 64

/* UART Buffer Defines */
#define UART0_RX_BUFFER_SIZE 64 /* 1,2,4,8,16,32,64,128 or 256 bytes */
#define UART0_RX_BUFFER_MASK ( UART0_RX_BUFFER_SIZE - 1 )
#if ( UART0_RX_BUFFER_SIZE & UART0_RX_BUFFER_MASK )
#error RX buffer size is not a power of 2
#endif
#define UART0_TX_BUFFER_SIZE 64 /* 1,2,4,8,16,32,64,128 or 256 bytes */
#define UART0_TX_BUFFER_MASK ( UART0_TX_BUFFER_SIZE - 1 )

#define UART1_RX_BUFFER_SIZE 64
#define UART1_RX_BUFFER_MASK ( UART1_RX_BUFFER_SIZE - 1 )
#define UART1_TX_BUFFER_SIZE 64
#define UART1_TX_BUFFER_MASK ( UART1_TX_BUFFER_SIZE - 1 )

/* Static Variables */
static unsigned char UART0_RxBuf[UART0_RX_BUFFER_SIZE];
static volatile unsigned char UART0_RxHead;
static volatile unsigned char UART0_RxTail;
static unsigned char UART0_TxBuf[UART0_TX_BUFFER_SIZE];
static volatile unsigned char UART0_TxHead;
static volatile unsigned char UART0_TxTail;

static unsigned char UART1_RxBuf[UART1_RX_BUFFER_SIZE];
static volatile unsigned char UART1_RxHead;
static volatile unsigned char UART1_RxTail;
static unsigned char UART1_TxBuf[UART1_TX_BUFFER_SIZE];
static volatile unsigned char UART1_TxHead;
static volatile unsigned char UART1_TxTail;

/* initialize UART */
/* ??? Assume that CPU_FREQ contains the correct value, otherwise just guess (not implemented yet :( )*/
/* also assuming 8n1 until I need to change this */
void UART_Init(char UART_Number, unsigned int baudrate)
{
   unsigned char x;
   if (UART_Number == 0)
   {
      outb(UBRR0H, (unsigned char)(baudrate >> 8));
      outb(UBRR0L, (unsigned char)(baudrate));
      /* enable UART receiver and transmitter, and receive interrupt */
      outb(UCSR0B, (1<<RXCIE0) | (1<<RXEN0) | (1<<TXEN0));
      /*   outp((1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0), UCSRC); */
      outb(UCSR0C, 0x86);
      x = 0; /* flush receive buffer */
      UART0_RxTail = x;
      UART0_RxHead = x;
      UART0_TxTail = x;
      UART0_TxHead = x;
   }
   else
   {
      outb(UBRR1H, (unsigned char)(baudrate >> 8));
      outb(UBRR1L, (unsigned char)(baudrate));
      /* enable UART receiver and transmitter, and receive interrupt */
      outb(UCSR1B, (1<<RXCIE1) | (1<<RXEN1) | (1<<TXEN1));
      /*   outp((1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0), UCSRC); */
      outb(UCSR1C, 0x86);
      x = 0; /* flush receive buffer */
      UART1_RxTail = x;
      UART1_RxHead = x;
      UART1_TxTail = x;
      UART1_TxHead = x;
   }
}

/* interrupt handlers */
//interrupt [UART_RX_vect] void UART_RX_interrupt( void )
SIGNAL(SIG_USART0_RECV)
{
   unsigned char data;
   unsigned char tmphead;
   data = inb(UDR0);
   /* calculate buffer index */
   tmphead = ( UART0_RxHead + 1 ) & UART0_RX_BUFFER_MASK;
   UART0_RxHead = tmphead; /* store new index */
   if ( tmphead == UART0_RxTail )
   {
      /* ERROR! Receive buffer overflow */
   }
   UART0_RxBuf[tmphead] = data; /* store received data in buffer */
}

//interrupt [UART_UDRE_vect] void UART_TX_interrupt(void)
SIGNAL(SIG_USART0_DATA)
{
   unsigned char tmptail;

   /* check if all data is transmitted */
   if ( UART0_TxHead != UART0_TxTail )
   {
      /* calculate buffer index */
      tmptail = ( UART0_TxTail + 1 ) & UART0_TX_BUFFER_MASK;
      UART0_TxTail = tmptail; /* store new index */
      outb(UDR0, UART0_TxBuf[tmptail]);
   }
   else
   {
      cbi(UCSR0B, UDRIE0);
   }
}

/* interrupt handlers */
//interrupt [UART_RX_vect] void UART_RX_interrupt( void )
SIGNAL(SIG_USART1_RECV)
{
   unsigned char data;
   unsigned char tmphead;
   data = inb(UDR1);
   /* calculate buffer index */
   tmphead = ( UART1_RxHead + 1 ) & UART1_RX_BUFFER_MASK;
   UART1_RxHead = tmphead; /* store new index */
   if ( tmphead == UART1_RxTail )
   {
      /* ERROR! Receive buffer overflow */
   }
   UART1_RxBuf[tmphead] = data; /* store received data in buffer */
}

//interrupt [UART_UDRE_vect] void UART_TX_interrupt(void)
SIGNAL(SIG_USART1_DATA)
{
   unsigned char tmptail;

   /* check if all data is transmitted */
   if ( UART1_TxHead != UART1_TxTail )
   {
      /* calculate buffer index */
      tmptail = ( UART1_TxTail + 1 ) & UART1_TX_BUFFER_MASK;
      UART1_TxTail = tmptail; /* store new index */
      outb(UDR1, UART1_TxBuf[tmptail]);
   }
   else
   {
      cbi(UCSR1B, UDRIE1);
   }
}


/* Write function to transmit 1 byte */
void UART_TransmitByte(char UART_Number, unsigned char Data)
{
   unsigned char tmphead;
   if (UART_Number == 0)
   {
      tmphead = ( UART0_TxHead + 1 ) & UART0_TX_BUFFER_MASK; /* wait for free space in buffer */
      while ( tmphead == UART0_TxTail );

      UART0_TxBuf[tmphead] = Data; /* store data in buffer */
      UART0_TxHead = tmphead; /* store new index */
      sbi(UCSR0B, UDRIE0);
   }
   else
   {
      tmphead = ( UART1_TxHead + 1 ) & UART1_TX_BUFFER_MASK; /* wait for free space in buffer */
      while ( tmphead == UART1_TxTail );

      UART1_TxBuf[tmphead] = Data; /* store data in buffer */
      UART1_TxHead = tmphead; /* store new index */
      sbi(UCSR1B, UDRIE1);
   }
}

/* Write function to transmit 2 bytes */
void UART_TransmitWord(char UART_Number, unsigned int Data)
{
   UART_TransmitByte(UART_Number, (Data >> 8) & 0xFF); /* send high byte */
   UART_TransmitByte(UART_Number, Data & 0xFF);        /* send low byte */
}

void UART_TransmitLine(char UART_Number, unsigned char *TransmitBuffer)
{
   unsigned char BufferCounter = 0;
   while (TransmitBuffer[BufferCounter] != '\0')
   {
      UART_TransmitByte(UART_Number, TransmitBuffer[BufferCounter++]);
   }
}

unsigned char UART_ReceiveByte(char UART_Number)
{
   unsigned char tmptail;
   if (UART_Number == 0)
   {
      while ( UART0_RxHead == UART0_RxTail ) ; /* wait for incomming data */
      tmptail = ( UART0_RxTail + 1 ) & UART0_RX_BUFFER_MASK;/* calculate buffer index */
      UART0_RxTail = tmptail; /* store new index */
      return UART0_RxBuf[tmptail]; /* return data */
   }
   else
   {
      while ( UART1_RxHead == UART1_RxTail ) ; /* wait for incomming data */
      tmptail = ( UART1_RxTail + 1 ) & UART1_RX_BUFFER_MASK;/* calculate buffer index */
      UART1_RxTail = tmptail; /* store new index */
      return UART1_RxBuf[tmptail]; /* return data */
   }
}

unsigned int UART_ReceiveWord(char UART_Number)
{
   unsigned int ReturnValue;
   ReturnValue = UART_ReceiveByte(UART_Number) << 8;
   ReturnValue |= UART_ReceiveByte(UART_Number);
   return(ReturnValue);
}

void UART_ReceiveLine(char UART_Number, unsigned char *ReturnBuffer) // ReturnBuffer must be at least UART_BUFFER_LENGTH long
{
   unsigned char BufferCounter = 0;
//   unsigned char OutputLine[20];
//   unsigned char TempLine[20];
   while (BufferCounter < UART_BUFFER_LENGTH - 2)
   {
      ReturnBuffer[BufferCounter] = UART_ReceiveByte(UART_Number);
      if ((ReturnBuffer[BufferCounter] == '\n') || (ReturnBuffer[BufferCounter] == '\r'))
      {
         ReturnBuffer[BufferCounter] = '\0';
         break;
      }
      if (((ReturnBuffer[BufferCounter] == '\b') || (ReturnBuffer[BufferCounter] == 127)) && (BufferCounter > 0))
      {
         BufferCounter --;
         UART_TransmitByte(UART_Number, '\b');
         continue;
      }
      UART_TransmitByte(UART_Number, ReturnBuffer[BufferCounter]);
      BufferCounter ++;
   }
   ReturnBuffer[BufferCounter++] = '\0';
}

unsigned char UART_Data(char UART_Number)
{
   if (UART_Number == 0)
   {
      if (UART0_RxHead != UART0_RxTail) return(1);
   }
   else if (UART_Number == 0)
   {
      if (UART1_RxHead != UART1_RxTail) return(1);
   }
   return(0);
//   return(UART_RxHead != UART_RxTail); /* return 0 (FALSE) if the receive buffer is empty */
}

void UART_clr(char UART_Number)
{ // Send a 'clear screen' to a VT100 terminal
   UART_TransmitByte(UART_Number, 0x1B);
   UART_TransmitByte(UART_Number, '[');
   UART_TransmitByte(UART_Number, 'H');
   UART_TransmitByte(UART_Number, 0x1B);
   UART_TransmitByte(UART_Number, '[');
   UART_TransmitByte(UART_Number, '2');
   UART_TransmitByte(UART_Number, 'J');
   UART_TransmitByte(UART_Number, 0x00);
}


#endif
